from __future__ import annotations

import copy
from typing import TYPE_CHECKING

from blacksquare.types import CellIndex, CellValue, Direction, SpecialCellValue

if TYPE_CHECKING:
    from blacksquare.crossword import Crossword
    from blacksquare.word import Word


class Cell:
    """An object representing a single cell in the crossword. Should not be constructed
    by the user.
    """

    def __init__(
        self,
        parent_crossword: Crossword,
        index: CellIndex,
        value: CellValue = SpecialCellValue.EMPTY,
    ):
        self._parent = parent_crossword
        self._index = index
        self._value = _parse_cell_input(value)
        self.shaded = False
        self.circled = False

    @property
    def parent_crossword(self) -> Crossword:
        return self._parent

    def get_parent_word(self, direction: Direction) -> Word:
        """Get the word to which the cell belongs in the given direction.

        Args:
            direction: The direction.

        Returns:
            The parent word.
        """
        return self._parent.get_word_at_index(self.index, direction)

    @property
    def value(self) -> CellValue:
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = _parse_cell_input(new_value)

    @property
    def index(self) -> CellIndex:
        return self._index

    @property
    def number(self) -> int | None:
        return self._parent.get_cell_number(self._index)

    @property
    def symmetric_image(self) -> Cell | list[Cell] | None:
        result = self.parent_crossword.get_symmetric_cell_index(self.index)
        if not result:
            return
        elif isinstance(result, list):
            return [self.parent_crossword[i] for i in result]
        else:
            return self.parent_crossword[result]

    @property
    def str(self) -> str:
        if isinstance(self._value, str):
            return self._value
        elif isinstance(self._value, SpecialCellValue):
            return self._value.str

    def __repr__(self):
        return f"Cell({repr(self._value)})"

    def __eq__(self, other):
        if isinstance(other, Cell):
            return self._value == other._value
        else:
            return self._value == other

    def __deepcopy__(self, memo):
        copied = copy.copy(self)
        copied._parent = None
        return copied


def _parse_cell_input(value: CellValue) -> CellValue:
    """Helper function to sanitize cell inputs.

    Args:
        value: The input value.

    Raises:
        ValueError: For invalid cell values.

    Returns:
        The cell value, either as a normalized string, or a SpecialCellValue enum.
    """
    if isinstance(value, SpecialCellValue):
        return value
    elif not isinstance(value, str) or len(value) != 1:
        raise ValueError
    else:
        if value in SpecialCellValue.BLACK.input_str_reprs:
            return SpecialCellValue.BLACK
        elif value in SpecialCellValue.EMPTY.input_str_reprs:
            return SpecialCellValue.EMPTY
        else:
            return value.upper()



from blacksquare.crossword import Crossword
from blacksquare.symmetry import Symmetry
from blacksquare.types import Direction, SpecialCellValue
from blacksquare.word_list import DEFAULT_WORDLIST, WordList

BLACK, EMPTY = SpecialCellValue.BLACK, SpecialCellValue.EMPTY
ACROSS, DOWN = Direction.ACROSS, Direction.DOWN

__all__ = [
    "Crossword",
    "Symmetry",
    "DEFAULT_WORDLIST",
    "WordList",
    "BLACK",
    "EMPTY",
    "ACROSS",
    "DOWN",
]



from pathlib import Path

import pytest

from blacksquare import ACROSS, BLACK, DOWN, EMPTY, Crossword, WordList
from blacksquare.symmetry import Symmetry


class TestCrosswordIndexing:
    def test_index_setter_word(self, xw):
        xw[ACROSS, 5] = "qZb"
        assert xw[ACROSS, 5].value == "QZB"
        assert xw[DOWN, 1].value == "BBZ"

    def test_index_setter_letter(self, xw):
        xw[2, 1] = "z"
        assert xw[ACROSS, 5].value == " Z "
        assert xw[DOWN, 1].value == "BBZ"

    def test_wrong_word_length(self, xw):
        for bad_val in ["AA", "AAAA"]:
            with pytest.raises(ValueError):
                xw[ACROSS, 5] = bad_val
        with pytest.raises(ValueError):
            xw[1, 0] = "AA"

    @pytest.mark.parametrize("bad_val", [["A", "A"], 72, Crossword(2)[ACROSS, 1]])
    def test_wrong_types(self, xw, bad_val):
        with pytest.raises(ValueError):
            xw[ACROSS, 5] = bad_val

    @pytest.mark.parametrize(
        "bad_index", [(ACROSS, 2), DOWN, (1, 1, 0), (1, slice(1, 2))]
    )
    def test_invalid_index(self, xw, bad_index):
        with pytest.raises(IndexError):
            xw[bad_index]
        with pytest.raises(IndexError):
            xw[bad_index] = "A"

    def test_unchecked_word_indexing(self):
        xw = Crossword(5, 5, symmetry=None)
        ind = [1, 2, 3]
        for i, j in zip(ind, ind[::-1]):
            xw[i, j] = BLACK
        xw[DOWN, 1] = "ABCDE"
        xw[DOWN, 1] = "     "
        xw.fill()


class TestParsing:
    def test_parsing_indices(self):
        xw = Crossword(5)
        BLACK_inds = [(0, 0), (0, 1), (2, 2), (4, 4), (4, 3)]
        for ind in BLACK_inds:
            xw[ind] = BLACK
        expected_keys = [
            (ACROSS, 1),
            (ACROSS, 4),
            (ACROSS, 6),
            (ACROSS, 7),
            (ACROSS, 8),
            (ACROSS, 10),
            (DOWN, 1),
            (DOWN, 2),
            (DOWN, 3),
            (DOWN, 4),
            (DOWN, 5),
            (DOWN, 9),
        ]
        assert set(expected_keys) == set([w.index for w in xw.iterwords()])

    def test_reparse_after_removing_black(self, xw):
        old_indices = set([w.index for w in xw.iterwords()])
        xw[0, 0] = " "
        assert set([w.index for w in xw.iterwords()]) != old_indices

    def test_clues_after_reparse(self, xw: Crossword):
        xw[ACROSS, 5].clue = "across to forget"
        xw[ACROSS, 4].clue = "across to keep"
        xw[DOWN, 4].clue = "down to keep"
        xw[DOWN, 2].clue = "down to forget"
        xw[2, 2] = BLACK
        assert xw[ACROSS, 3].clue == "across to keep"
        assert xw[DOWN, 3].clue == "down to keep"
        for k, c in xw.clues.items():
            if k not in ((ACROSS, 3), (DOWN, 3)):
                assert c == ""

    def test_single_letter(self):
        grid = [
            ["#", "#", "#", "#", "A"],
            ["B", "A", "D", "#", "C"],
            ["#", "#", "#", "#", "T"],
        ]
        xw = Crossword(grid=grid)
        assert xw[ACROSS, 2].value == "BAD"


class TestConversion:
    def test_to_puz(self, xw, tmp_path):
        filename = tmp_path / "test.puz"
        xw.to_puz(filename)
        assert tmp_path.exists()
        loaded = Crossword.from_puz(filename)
        assert all(x == y for x, y in zip(loaded.itercells(), xw.itercells()))

    def test_from_puz(self):
        Crossword.from_puz(Path(__file__).parent / "dummy.puz")

    def test_to_pdf(self, xw, tmp_path):
        filename = tmp_path / "test.pdf"
        xw.to_pdf(filename, ["Line 1", "Line 2"])
        assert tmp_path.exists()


class TestCrosswordProperties:
    def test_symmetric_cell_index(self):
        xw = Crossword(5, 7)
        assert xw.get_symmetric_cell_index((0, 1)) == (4, 5)
        assert xw.get_symmetric_cell_index((2, 3)) == (2, 3)

    def test_num_rows_cols(self):
        xw = Crossword(5, 7)
        assert xw.num_rows == 5
        assert xw.num_cols == 7

    def test_iterwords(self, xw):
        across_indices = [(ACROSS, n) for n in [1, 4, 5]]
        down_indices = [(DOWN, n) for n in [1, 2, 3, 4]]
        assert list(w.index for w in xw.iterwords()) == across_indices + down_indices
        assert list(w.index for w in xw.iterwords(ACROSS)) == across_indices
        assert list(w.index for w in xw.iterwords(DOWN)) == down_indices

    def test_get_indices(self, xw):
        assert xw.get_indices((ACROSS, 4)) == [(1, 0), (1, 1), (1, 2), (1, 3)]

    def test_get_word_at_index(self, xw: Crossword):
        assert xw.get_word_at_index((1, 1), ACROSS) == xw[ACROSS, 4]
        assert xw.get_word_at_index((1, 1), DOWN) == xw[DOWN, 1]
        assert xw.get_word_at_index((0, 0), ACROSS) is None

    def test_set_word(self, xw):
        xw.set_word((ACROSS, 1), "DDD")
        assert xw[ACROSS, 1].value == "DDD"

    def test_set_cell(self, xw):
        xw.set_cell((0, 1), "D")
        assert xw[ACROSS, 1].value == "DCD"

    def test_copy(self, xw):
        xw2 = xw.copy()
        assert xw2 is not xw
        xw2.set_cell((2, 1), "E")
        assert xw2[ACROSS, 5].value == " E "
        assert xw[ACROSS, 5].value == "   "

    def test_subgrids(self, xw: Crossword):
        graphs = xw.get_disconnected_open_subgrids()
        assert len(graphs) == 1
        assert len(graphs[0]) == 4
        xw[2, 0] = "Z"
        xw[1, 3] = " "
        graphs = xw.get_disconnected_open_subgrids()
        assert len(graphs) == 2


class TestCrosswordFill:
    def test_fill(self, xw):
        solution = xw.fill()
        assert solution[ACROSS, 5].value == "ABB"
        assert len(set(w.value for w in solution.iterwords())) == len(
            list(solution.iterwords())
        )

    def test_fill_with_repeats(self, xw):
        solution = xw.fill(allow_repeats=True)
        assert solution[ACROSS, 5].value == "BBB"

    def test_find_solutions_with_custom_dictionary(self, xw):
        solution = xw.fill(
            word_list=WordList(
                {
                    "BBB": 1,
                    "AC": 1,
                    "CBX": 1,
                    "CBA": 1,
                    "AA": 0.1,
                    "AB": 0.1,
                    "CCX": 1,
                    "CCA": 0.1,
                }
            ),
        )
        assert solution[ACROSS, 5].value == "CBX"

    def test_unsolvable(self, xw):
        xw[1, 0] = "J"
        xw[1, 1] = EMPTY
        xw[0, 1] = "Q"
        solution = xw.fill()
        assert solution is None

    def test_unsolvable_if_repeats_needed(self):
        xw = Crossword(3)
        word_list = WordList({"AAA": 1})
        assert xw.fill(word_list) is None

    def test_unsolvable_with_score_filter(self, xw):
        assert xw.fill(score_filter=1.0) is None

    def test_filled_by_setting_letter(self):
        xw = Crossword(5, symmetry=None)
        for i in range(5):
            for j in range(4):
                xw[i, j] = BLACK
            xw[i, 4] = "A"
        result = xw.fill()
        assert result is not None


def test_symmetry_requirements():
    with pytest.raises(ValueError):
        Crossword(5, 7, symmetry=Symmetry.FULL)



import pytest

from blacksquare import ACROSS, DOWN, Crossword, WordList
from blacksquare.word_list import MatchWordList


class TestWordList:
    def test_from_list(self):
        words = ["ABC", "AAA", "ABCD", "ABCDE"]
        word_list = WordList(words)
        assert word_list.words == sorted(words)
        assert all([s == 1.0 for s in word_list.scores])

    def test_from_dict(self):
        word_dict = {"aaa": 10, "bBB": 100, "C CC": 50}
        word_list = WordList(word_dict)
        assert word_list.words == ["BBB", "CCC", "AAA"]
        assert word_list.scores == [1.0, 0.5, 0.1]

    def test_from_file(self, tmp_path):
        dict_file = tmp_path / "wordlist.dict"
        with dict_file.open("w") as f:
            f.writelines(["aaa;10\n", "bBB;100\n", "C CC;50\n"])
        word_list = WordList(dict_file)
        assert word_list.words == ["BBB", "CCC", "AAA"]
        assert word_list.scores == [1.0, 0.5, 0.1]

    @pytest.mark.parametrize("blank_string", ["?", " ", "_"])
    def test_find_matches_str(self, word_list, blank_string):
        matches = word_list.find_matches_str(f"{blank_string}BB")
        assert matches.words == ["ABB", "BBB"]

    @pytest.mark.parametrize("no_match_str", ["ZZZ", "Z??", "ABCDABCD?"])
    def test_no_matches(self, word_list, no_match_str):
        assert len(word_list.find_matches_str(no_match_str)) == 0

    def test_find_matches(self, xw, word_list):
        across_matches = word_list.find_matches(xw[ACROSS, 5])
        assert across_matches.words == ["ABB", "ABC", "CCB", "BBB", "BCD"]

        down_matches = word_list.find_matches(xw[DOWN, 1])
        assert down_matches.words == ["BBB"]

    def test_find_matches_filled(self, xw: Crossword, word_list: WordList):
        matches = word_list.find_matches(xw[ACROSS, 1])
        assert matches.words == ["BCD"]

    def test_get_score(self, word_list):
        assert word_list.get_score("BBB") == 0.1
        assert word_list.get_score("XYZ") is None

    def test_add_word_list(self):
        word_list_1 = {"ZZZ": 0.1, "AAA": 1.0}
        word_list_2 = {"ABC": 0.9, "XYZ": 1.0}
        new_word_list = WordList(word_list_1) + WordList(word_list_2)
        assert new_word_list.words == ["AAA", "XYZ", "ABC", "ZZZ"]
        assert new_word_list.scores == [1.0, 1.0, 0.9, 0.1]

    def test_score_filter(self, word_list):
        assert len(word_list.score_filter(0.5)) == 6

    def test_serialization(self, tmp_path):
        words = """
        AAA;1.0
        BBB;0.99
        BB;0.5
        C;0.1
        """
        with (tmp_path / "list.dict").open("w") as f:
            f.write(words)

        xw = Crossword(3)
        wl = WordList(tmp_path / "list.dict")
        dict_matches = wl.find_matches(xw[ACROSS, 1])
        wl.to_npz(tmp_path / "list.npz")
        wl_from_npz = WordList(tmp_path / "list.npz")
        npz_matches = wl_from_npz.find_matches(xw[ACROSS, 1])
        assert dict_matches.words == npz_matches.words


class TestMatchWordList:
    @pytest.fixture
    def matches(self, word_list) -> MatchWordList:
        return word_list.find_matches_str("???")

    def test_letter_scores_at_index(self, matches: MatchWordList):
        assert matches.letter_scores_at_index(1) == {"B": 2.1, "C": 1.1}

    def test_rescore(self, matches: MatchWordList):
        def rescore_fn(word, score):
            if "A" in word:
                return 2.0 * score
            else:
                return 1.0 * score

        rescored = matches.rescore(rescore_fn)
        assert rescored.get_score("ABC") == 2.0
        assert rescored.get_score("BCD") == 0.1

    def test_filter_words(self, matches: MatchWordList):
        assert len(matches) == 5
        assert len(matches.filter_words(["ABC", "ABB"])) == 3



from typing import Any

import numpy as np


def sum_by_group(groups: np.ndarray, values: np.ndarray) -> dict:
    """A utility function to perform grouped summing of values over numpy arrays.

    Args:
        groups: An array describing what group each value corresponds to.
        values: An array of values.

    Returns:
        A dictionary mapping each unique group the sum of its values.
    """
    sort_indices = np.argsort(groups)
    sorted_groups = groups[sort_indices]
    sorted_values = values[sort_indices]
    unique_groups, unique_indices = np.unique(sorted_groups, return_index=True)
    group_values = np.split(sorted_values, unique_indices[1:])
    return {g: v.sum() for g, v in zip(unique_groups, group_values)}


def is_intlike(x: Any) -> bool:
    """A helper function for doing type checking on possibly-numpy integers.

    Args:
        x: The input value.

    Returns:
        True if the number is a python integer or a numpy integer.
    """
    return isinstance(x, (int, np.integer))



CSS_TEMPLATE = """
.crossword{suffix} {{
    display: grid;
    grid-template-columns: repeat({num_cols}, 1fr);
    grid-auto-rows: 1fr;
    gap: 0px;
    width: {width}px;
    height: {height}px;
}}

.crossword-cell{suffix} {{
    position: relative;
    background-color: white;
    outline: 1px solid black;
    font-family: Arial, Helvetica, sans-serif;
    aspect-ratio: 1;
}}

.crossword-cell{suffix} .cell-number {{
    position: absolute;
    top: 2px;
    left: 2px;
    font-size: {num_font_size}px;
    color: black;
    user-select: none;
}}

.crossword-cell{suffix} .letter {{
    font-size: {val_font_size}px;
    position: absolute;
    bottom: 0;
    left: 50%;
    transform: translate(-50%, 0%);
    color: black;
}}

.black {{
    background-color: black;
    outline: 1px solid gray;
}}

.gray {{
    background-color: lightgray;
}}

.crossword-cell{suffix} .circle {{
    position: absolute;
    border-radius: 50%;
    border: 1px solid black;
    height: {circle_dim}px;
    width: {circle_dim}px;
    margin: -1px;
}}
"""



from __future__ import annotations

from enum import Enum


class Direction(Enum):
    """An Enum representing the directions of words in a crossword."""

    ACROSS = "Across"
    DOWN = "Down"

    @property
    def opposite(self) -> Direction:
        if self == Direction.ACROSS:
            return Direction.DOWN
        else:
            return Direction.ACROSS

    def __lt__(self, other) -> bool:
        if isinstance(other, Direction):
            return self == Direction.ACROSS and other == Direction.DOWN
        return NotImplemented

    def __repr__(self):
        return f"<{self.value}>"


class SpecialCellValue(Enum):
    "An enum representing blank and empty cell values in a crossword."

    BLACK = "Black"
    EMPTY = "Empty"

    @property
    def input_str_reprs(self) -> list[str]:
        if self == SpecialCellValue.BLACK:
            return [".", "#"]
        elif self == SpecialCellValue.EMPTY:
            return [" ", "-", "?", "_"]

    @property
    def str(self) -> str:
        if self == SpecialCellValue.BLACK:
            return "█"
        elif self == SpecialCellValue.EMPTY:
            return " "

    def __repr__(self):
        return f"<{self.value}>"


ACROSS = Direction.ACROSS
DOWN = Direction.DOWN

WordIndex = tuple[Direction, int]
CellIndex = tuple[int, int]
CellValue = str | SpecialCellValue



from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import numpy as np

from blacksquare.types import Direction, SpecialCellValue, WordIndex
from blacksquare.word_list import INVERSE_CHARACTER_FREQUENCIES, MatchWordList, WordList

if TYPE_CHECKING:
    from blacksquare.cell import Cell
    from blacksquare.crossword import Crossword


class Word:
    """An object representing a single Word, with awareness of the parent grid. Should
    not be constructed by the user.
    """

    def __init__(
        self,
        parent_crossword: Crossword,
        direction: Direction,
        number: int,
        clue: str = "",
    ):
        """Returns a new Word object.

        Args:
            parent_crossword: The parent crossword for the word.
            direction: The direction of the word.
            number: The ordinal number of the word.
            clue: The clue associated with the word. Defaults to "".
        """
        self.clue = clue
        self._parent = parent_crossword
        self._direction = direction
        self._number = number

    def __getitem__(self, key) -> Cell:
        return self.cells[key]

    def __setitem__(self, key, value):
        self.cells[key].value = value

    @property
    def direction(self) -> Direction:
        """Direction: The direction of the word."""
        return self._direction

    @property
    def number(self) -> int:
        """The number of the word."""
        return self._number

    @property
    def index(self) -> WordIndex:
        """The (direction, number) index of the word."""
        return (self.direction, self.number)

    @property
    def cells(self) -> list[Cell]:
        return self._parent.get_word_cells(self.index)

    @property
    def value(self) -> str:
        # TODO: rename to str
        """The current fill value of the word"""
        return "".join([c.str for c in self.cells])

    # Todo: array, str?

    @value.setter
    def value(self, new_value: str):
        self._parent[self.index] = new_value

    def is_open(self) -> bool:
        """Does the word contain any blank spaces.

        Returns:
            True if any of the letters are blank.
        """
        return np.equal(self.cells, SpecialCellValue.EMPTY).any()

    @property
    def crosses(self) -> list[Word]:
        """Returns the words that cross the current word.

        Returns:
            A list of Word objects corresponding to the crosses.
        """
        return [cell.get_parent_word(self.direction.opposite) for cell in self.cells]

    @property
    def symmetric_image(self) -> Word | list[Word] | None:
        result = self._parent.get_symmetric_word_index(self.index)
        if not result:
            return
        elif isinstance(result, list):
            return [self._parent[i] for i in result]
        else:
            return self._parent[result]

    def find_matches(
        self, word_list: WordList | None = None, allow_repeats: bool = False
    ) -> MatchWordList:
        """Finds matches for the word, ranking matches by how many valid crosses they
        allow.

        Args:
            word_list: The word list to use for matching. If None, the default wordlist
                of the parent crossword is used.
            allow_repeats: Whether to include words that are already in the grid.
                Defaults to False.

        Returns:
            The matching words, scored by compatible crosses.
        """
        word_list = self._parent.word_list if word_list is None else word_list
        self_len = len(self)
        matches = word_list.find_matches(self)
        if not allow_repeats:
            matches = matches.filter_words(
                [
                    w.value
                    for w in self._parent.iterwords()
                    if len(w) == self_len and not w.is_open()
                ]
            )
        open_indices = np.argwhere(
            np.equal(self.cells, SpecialCellValue.EMPTY)
        ).squeeze(axis=1)
        letter_scores_per_index = {}
        for idx in open_indices:
            cross = self.crosses[idx]
            if cross is None:
                continue
            cross_index = cross.crosses.index(self)
            cross_matches = word_list.find_matches(cross)
            letter_scores_per_index[idx] = cross_matches.letter_scores_at_index(
                cross_index
            )

        def score_word_fn(word: str, score: float) -> float:
            per_letter_scores = [
                letter_scores_per_index[i].get(word[i], 0)
                * INVERSE_CHARACTER_FREQUENCIES.get(word[i], 1)
                for i in open_indices
                if i in letter_scores_per_index
            ]
            return np.log(np.prod(per_letter_scores) + 1) * score

        return matches.rescore(score_word_fn)

    def __repr__(self):
        return f'Word({self.direction.value} {self.number}: "{self.value.replace(" ", "?")})"'

    def __len__(self):
        return len(self.cells)

    def __deepcopy__(self, memo):
        copied = copy.copy(self)
        copied._parent = None
        return copied



import numpy as np
import pytest

from blacksquare.symmetry import Symmetry


@pytest.fixture
def rect():
    return np.arange(35).reshape(5, 7)


@pytest.fixture
def square():
    return np.arange(49).reshape(7, 7)


def test_rotation(rect):
    image = Symmetry.ROTATIONAL.apply(rect).grid
    assert rect[0, 1] == image[4, 5]
    assert rect[2, 3] == image[2, 3]


def test_full(square):
    images = [x.grid for x in Symmetry.FULL.apply(square)]
    assert set(i[0, 1] for i in images) == {
        square[1, 0],
        square[5, 6],
        square[6, 5],
        square[0, 5],
        square[5, 0],
        square[6, 1],
        square[1, 6],
    }
    assert set(i[0, 3] for i in images) == {
        square[0, 3],
        square[3, 0],
        square[3, 6],
        square[6, 3],
    }
    assert set(i[3, 3] for i in images) == {square[3, 3]}


def test_vertical(rect):
    image = Symmetry.VERTICAL.apply(rect).grid
    assert rect[0, 1] == image[0, 5]
    assert np.all(rect[:, 3] == image[:, 3])


def test_horizontal(rect):
    image = Symmetry.HORIZONTAL.apply(rect).grid
    assert rect[0, 1] == image[4, 1]
    assert np.all(rect[2, :] == image[2, :])


def test_biaxial(rect):
    images = [x.grid for x in Symmetry.BIAXIAL.apply(rect)]
    assert set(i[0, 1] for i in images) == {rect[4, 1], rect[0, 5], rect[4, 5]}
    assert set(i[0, 3] for i in images) == {rect[0, 3], rect[4, 3]}
    assert set(i[2, 3] for i in images) == {rect[2, 3]}


def test_ne_diagonal(square):
    image = Symmetry.NE_DIAGONAL.apply(square).grid
    assert square[0, 1] == image[5, 6]
    assert square[0, 6] == image[0, 6]


def test_nw_diagonal(square):
    image = Symmetry.NW_DIAGONAL.apply(square).grid
    assert square[0, 1] == image[1, 0]
    assert square[2, 2] == image[2, 2]
